@@ -687,7 +687,10 @@ class ChainlitAdapter(SDKAdapter):
             if is_debug_mode_active:
                 current_step.name = tool_name
                 if is_sub_agent_execution:
-                    current_step.input = tool_call.payload['bodyParams']['input_task']
+                    input_task_str = tool_call.payload['bodyParams'].get("input_task", "")
+                    if not input_task_str:
+                        input_task_str = json.dumps(tool_call.payload)
+                    current_step.input = input_task_str
                 else:
                     current_step.input = tool_call.payload
             
