@@ -1,12 +1,18 @@
 import json
-from typing import Dict, Optional, Callable, Union, List
-from xpander_sdk import ToolCall, ToolCallType
-from .base import SDKAdapter
+from typing import Dict, Optional, Callable, Union, List, Any
+from xpander_sdk import ToolCall, ToolCallType, LLMTokens, Tokens
+from .base import SDKAdapter, AGENT_FINISH_TOOL_ID
 from llama_index.core.tools import FunctionTool
 from llama_index.core.agent import AgentRunner
 from llama_index.core.prompts import PromptTemplate
 from llama_index.core.memory.types import ChatMessage
 from llama_index.core.llms import MessageRole
+from llama_index.core.callbacks import CallbackManager
+from llama_index.core.callbacks.pythonically_printing_base_handler import (
+    PythonicallyPrintingBaseHandler,
+)
+from llama_index.core.callbacks.schema import CBEventType, EventPayload
+
 from ...generic import generate_tool_call_id
 
 class LlamaIndexAdapter(SDKAdapter):
@@ -20,7 +26,7 @@ class LlamaIndexAdapter(SDKAdapter):
         agent (Agent): The xpander.ai agent instance.
     """
 
-    def __init__(self, api_key: str, agent_id: str, base_url: Optional[str] = None, organization_id: Optional[str] = None):
+    def __init__(self, api_key: str, agent_id: str, base_url: Optional[str] = None, organization_id: Optional[str] = None, with_metrics_report: Optional[bool] = False):
         """
         Initialize the LlamaIndexAdapter.
 
@@ -29,9 +35,11 @@ class LlamaIndexAdapter(SDKAdapter):
             agent_id (str): The ID of the xpander.ai agent.
             base_url (Optional[str], optional): The base URL for the xpander.ai API. Defaults to None.
             organization_id (Optional[str], optional): The organization ID, if applicable. Defaults to None.
+            with_metrics_report (Optional[bool], optional): If to auto-report metrics (llm & execution). Default to False.
         """
-        super().__init__(api_key, agent_id, base_url, organization_id)
+        super().__init__(api_key=api_key, agent_id=agent_id, base_url=base_url, organization_id=organization_id, with_metrics_report=with_metrics_report)
         self.agent.disable_agent_end_tool()  # No need since LlamaIndex handles it.
+        self.callback_manager = CallbackManager([LLMHandler(xpander=self)])
 
     def get_tools(self, return_as_function_tool: bool = True) -> List[Union[FunctionTool, Callable]]:
         """
@@ -104,7 +112,7 @@ class LlamaIndexAdapter(SDKAdapter):
 
                     # Convert result to string
                     stringified_result = json.dumps(xpander_agent_tool_invocation.result)
-
+                    
                     if not xpander_agent_tool_invocation.is_success:
                         raise Exception(f"Error running tool: {stringified_result}")
 
@@ -178,3 +186,58 @@ class LlamaIndexAdapter(SDKAdapter):
 
             if messages_to_append:
                 self.agent.add_messages(messages=messages_to_append)
+
+class LLMHandler(PythonicallyPrintingBaseHandler):
+    def __init__(self, xpander: LlamaIndexAdapter) -> None:
+        super().__init__(
+            event_starts_to_ignore=[], event_ends_to_ignore=[]
+        )
+        
+        self.xpander = xpander
+
+
+    def start_trace(self, trace_id: Optional[str] = None) -> None:
+        pass
+
+    def end_trace(
+        self,
+        trace_id: Optional[str] = None,
+        trace_map: Optional[Dict[str, List[str]]] = None,
+    ) -> None:
+        return
+
+    def on_event_start(
+        self,
+        event_type: CBEventType,
+        payload: Optional[Dict[str, Any]] = None,
+        event_id: str = "",
+        parent_id: str = "",
+        **kwargs: Any,
+    ) -> str:
+        pass
+
+    def on_event_end(
+        self,
+        event_type: CBEventType,
+        payload: Optional[Dict[str, Any]] = None,
+        event_id: str = "",
+        **kwargs: Any,
+    ) -> None:
+        if self.xpander.with_metrics_report:
+            if event_type == CBEventType.LLM and payload is not None:
+                # calculate total tokens
+                execution_id = self.xpander.agent.execution.id
+                llm_response = payload['response'].raw.model_dump()
+                tokens = LLMTokens(completion_tokens=llm_response['usage']['completion_tokens'],prompt_tokens=llm_response['usage']['prompt_tokens'],total_tokens=llm_response['usage']['total_tokens'])
+                if not execution_id in self.xpander.execution_tokens:
+                    self.xpander.execution_tokens[execution_id] = Tokens(worker=tokens)
+                else:
+                    self.xpander.execution_tokens[execution_id].worker.completion_tokens += tokens.completion_tokens
+                    self.xpander.execution_tokens[execution_id].worker.prompt_tokens += tokens.prompt_tokens
+                    self.xpander.execution_tokens[execution_id].worker.total_tokens += tokens.total_tokens
+                
+                # report llm usage
+                self.xpander.agent.report_llm_usage(llm_response=llm_response,llm_inference_duration=0)
+            elif event_type == CBEventType.AGENT_STEP and "response" in payload and payload["response"].response:
+                self.xpander.update_execution_result(is_success=True,result=payload["response"].response)
+                self.xpander.agent.report_execution_metrics(llm_tokens=self.xpander.execution_tokens[self.xpander.agent.execution.id],ai_model="N/A")
\ No newline at end of file
